#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "lnm/loop.h"
#include "lnm/loop_internal.h"

void lnm_loop_conn_io_req(lnm_loop *l, lnm_loop_conn *conn) {
  do {
    // Move remaining data to front of buffer
    memmove(conn->r.buf, &conn->r.buf[conn->r.read],
            conn->r.size - conn->r.read);
    conn->r.size -= conn->r.read;
    conn->r.read = 0;

    ssize_t res;
    size_t cap = LNM_LOOP_BUF_SIZE - conn->r.size;

    do {
      res = read(conn->fd, &conn->r.buf[conn->r.size], cap);
    } while (res < 0 && errno == EINTR);

    // Read can't be performed without blocking; we come back later
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    if (res <= 0) {
      conn->state = lnm_loop_state_end;

      return;
    }

    conn->r.size += res;
    l->data_read(conn);
  } while (conn->state == lnm_loop_state_req_io);
}

void lnm_loop_conn_io_res(lnm_loop *l, lnm_loop_conn *conn) {
  do {
    l->data_write(conn);

    ssize_t res;

    do {
      // Send with MSG_NOSIGNAL prevents closed pipes from exiting the program
      // with SIGPIPE
      res = send(conn->fd, conn->w.buf, conn->w.size, MSG_NOSIGNAL);
    } while (res < 0 && errno == EINTR);

    // Write can't be performed without blocking; we come back later
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    if (res < 0) {
      conn->state = lnm_loop_state_end;

      return;
    }

    // Move remaining data to front of buffer. Doing this here gives the data
    // writer function more space to work with
    memmove(conn->w.buf, &conn->w.buf[res], conn->w.size - res);
    conn->w.size -= res;
  } while (conn->state == lnm_loop_state_res_io);
}

void lnm_loop_conn_advance(lnm_loop *l, lnm_loop_conn *conn) {
  do {

    switch (conn->state) {
    case lnm_loop_state_req_io:
      lnm_loop_conn_io_req(l, conn);
      break;
    case lnm_loop_state_res_io:
      lnm_loop_conn_io_res(l, conn);
      break;
    case lnm_loop_state_req_work:
      do {
        l->data_read(conn);
      } while (conn->state == lnm_loop_state_req_work);
      break;
    case lnm_loop_state_res_work:
      do {
        l->data_write(conn);
      } while (conn->state == lnm_loop_state_res_work);
      break;
    default:;
    }
  }
  // Execute all blocking work if we're running in single-threaded mode
  while (l->wq == NULL && (conn->state == lnm_loop_state_req_work ||
                           conn->state == lnm_loop_state_res_work));
}
